/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.lucene.tests.mockfile;

import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Objects;

/**
 * A {@code FilterInputStream2} contains another {@code InputStream}, which it uses as its basic
 * source of data, possibly transforming the data along the way or providing additional
 * functionality.
 *
 * <p>Note: unlike {@link FilterInputStream} this class delegates every method by default. This
 * means to transform {@code read} calls, you need to override multiple methods. On the other hand,
 * it is less trappy: a simple implementation that just overrides {@code close} will not force bytes
 * to be read one-at-a-time.
 */
public class FilterInputStream2 extends InputStream {

  /** The underlying {@code InputStream} instance. */
  protected final InputStream delegate;

  /**
   * Construct a {@code FilterInputStream2} based on the specified base stream.
   *
   * <p>Note that base stream is closed if this stream is closed.
   *
   * @param delegate specified base stream.
   */
  public FilterInputStream2(InputStream delegate) {
    this.delegate = Objects.requireNonNull(delegate);
  }

  @Override
  public int read() throws IOException {
    return delegate.read();
  }

  @Override
  public int read(byte[] b) throws IOException {
    return delegate.read(b);
  }

  @Override
  public int read(byte[] b, int off, int len) throws IOException {
    return delegate.read(b, off, len);
  }

  @Override
  public long skip(long n) throws IOException {
    return delegate.skip(n);
  }

  @Override
  public int available() throws IOException {
    return delegate.available();
  }

  @Override
  public void close() throws IOException {
    delegate.close();
  }

  @Override
  public synchronized void mark(int readlimit) {
    delegate.mark(readlimit);
  }

  @Override
  public synchronized void reset() throws IOException {
    delegate.reset();
  }

  @Override
  public boolean markSupported() {
    return delegate.markSupported();
  }
}
